/*!
 * @file Adafruit_ZeroTimer.cpp
 *
 * @mainpage Adafruit ZeroTimer
 *
 * @section intro_sec Introduction
 *
 * Simple wrappers for TC modules 3,4,5 on SAMD21 and SAMD51
 * @section license License
 *
 * BSD license, all text above must be included in any redistribution
 *
 * 04/04/22 : Modified (partial POSIX adaptation, for jumper only) by Guénal Davalan
 */
#include "Adafruit_ZeroTimer.h"
#include "posix.h"
#include <unistd.h>

// FAKE define
#define GCM_TCC2_TC3 0
#define GCLK_CLKCTRL_ID 0
#define GCM_TC4_TC5 0
#define PM_APBCMASK_TC3 0
#define PM_APBCMASK_TC4 0
#define PM_APBCMASK_TC5 0
#define TC_CTRLA_SWRST 0
#define TC_STATUS_SLAVE 0
#define TC_CTRLA_ENABLE 0
#define TC_INTENCLR_MASK 0
#define TC_INTFLAG_MASK 0
#define TC_CTRLA_ENABLE 0
#define TC3 0


Tc tcArray[4];

// mostly from asfdoc_sam0_tc_basic_use_case.html

#ifndef TC_INST_MAX_ID
#define TC_INST_MAX_ID 5 ///< The max number of TC instances
#endif

#ifndef TC_INSTANCE_OFFSET
#define TC_INSTANCE_OFFSET 3 ///< The number of the first TC instance

#endif

extern "C" {

/**! togglebits for callbacks */
#define TC_CALLBACK_BITS 6
/**! How many callbacks we get */
#define TC_MAX_CALLBACKS (TC_CALLBACK_BITS * 3)

//static void (*__cb[TC_MAX_CALLBACKS])(void);
/** Bit mask for callbacks registered */
//static uint32_t _register_callback_mask = 0;
/** Bit mask for callbacks enabled */
//static uint32_t _enable_callback_mask = 0;
};

/*
static inline bool tc_is_syncing(Tc *const hw) {
#if defined(__SAMD51__)
  return hw->COUNT8.SYNCBUSY.reg > 0;
#else
  return (hw->COUNT8.STATUS.reg & TC_STATUS_SYNCBUSY);
#endif
}
*/

/**************************************************************************/
/*!
    @brief  Instantiate a ZeroTimer class
    @param  timernum The timer we are wrapping, 3 for TC3, 4 for TC4, etc!
*/
/**************************************************************************/
Adafruit_ZeroTimer::Adafruit_ZeroTimer(uint8_t timernum) {
  _timernum = timernum;
}

/**************************************************************************/
/*!
    @brief  Initializer for timer counter object
    @return True if we were able to init the timer, False on any failure
*/
/**************************************************************************/
bool Adafruit_ZeroTimer::tc_init() {
  return true;
}

/**************************************************************************/
/*!
    @brief  Use the TC to output a PWM signal on a given pin. Check datasheet
    to verify what pins can be connected to which TC output and what channel
    to use.
    @param pwmout True to enable the output, False to disable
    @param channum Which channel, 0 or 1, we want to output on.
    @param pin The Arduino pin name we want to have the PWM output on
    @returns True on success
*/
/**************************************************************************/
boolean Adafruit_ZeroTimer::PWMout(boolean pwmout, uint8_t channum,
                                   uint8_t pin) {
  return tc_init();
}

/**************************************************************************/
/*!
    @brief Whether or not to invert the output PWM
    @param invert True to invert, False to use default
*/
/**************************************************************************/
void Adafruit_ZeroTimer::invertWave(uint8_t invert) {
  //_waveform_invert_output = invert;

  tc_init();
}

/**************************************************************************/
/*!
    @brief Configure how we want to use the timer, based on ASF constants
    @param prescale What divider we want, e.g. TC_CLOCK_PRESCALER_DIV16 see
   https://asf.microchip.com/docs/latest/thirdparty.wireless.avr2130_lwmesh.apps.wsndemo.saml21_xpro_b_rf233/html/group__asfdoc__sam0__tc__group.html#ga98aed17b995157e67b9322a45f0ed5f4
    @param countersize Can be TC_COUNTER_SIZE_8BIT, TC_COUNTER_SIZE_16BIT or
   TC_COUNTER_SIZE_32BIT
    @param wavegen Can be TC_WAVE_WAVEGEN_NFRQ, TC_WAVE_WAVEGEN_MFRQ,
   TC_WAVE_WAVEGEN_NPWM or TC_WAVE_WAVEGEN_MPWM
    @param countdir Can be TC_COUNT_DIRECTION_UP or TC_COUNT_DIRECTION_DOWN
*/
/**************************************************************************/
void Adafruit_ZeroTimer::configure(tc_clock_prescaler prescale,
                                   tc_counter_size countersize,
                                   tc_wave_generation wavegen,
                                   tc_count_direction countdir) {
  _clock_prescaler = prescale;							   
  _counter_size = countersize;
  _wave_generation = wavegen;
  _count_direction = countdir;
  
  //fprintf(stderr, "Adafruit_ZeroTimer::configure : prescale:%d countersize:%d wavegen:%d countdir:%d\n", 
	//prescale, countersize, wavegen, countdir);
  
  tc_init();
}

/**************************************************************************/
/*!
    @brief Set up channel 0 of the timer for the 'top' or period setting value,
   and channel 1 for the 'compare' value
    @param period The period register value to use (you'll need to calculate
   this from the frequency desired)
    @param match The compare value, should be less than the period, determines
   the duty cycle
    @param channum Only for 8-bit counter mod, we can have the period set and
   *two* channel outputs
*/
/**************************************************************************/
void Adafruit_ZeroTimer::setPeriodMatch(uint32_t period, uint32_t match,
                                        uint8_t channum) {
}

/**************************************************************************/
/*!
    @brief Have a function called whenever the timer interrupt goes off
    @param enable True to enable the callback, False to disable
    @param cb_type Which channel to use, can be TC_CALLBACK_CC_CHANNEL0 or
   TC_CALLBACK_CC_CHANNEL1
    @param callback_func A function with no params or return, that will be
   called
*/
/**************************************************************************/
void Adafruit_ZeroTimer::setCallback(boolean enable, tc_callback cb_type,
                                     void (*callback_func)(void)) {
  
  //fprintf(stderr, "Adafruit_ZeroTimer::setCallback enable:%d cb_type:%d callback_func:%p\n",
	//enable, cb_type, callback_func);
	
  if (callback_func == NULL)
    return;

	tcArray[_timernum].callback=callback_func;
}

/**************************************************************************/
/*!
    @brief Set the timer counter's channel compare register to a value
    @param channum Which channel to use, can be 0 or 1
    @param compare The compare value to set the channel compare value to,
    will be cast to whatever size the timer is setup for (8/16/32 bit)
*/
/**************************************************************************/
void Adafruit_ZeroTimer::setCompare(uint8_t channum, uint32_t compare) {
  
  //fprintf(stderr, "Adafruit_ZeroTimer::setCompare channum:%d compare:%d\n", channum, compare);
  
  if (channum > 1)
    return;

  tcArray[_timernum].period=(20000L*compare)/920;
}

/**************************************************************************/
/*!
    @brief  Enable or disable the timer. Won't do anything if the status is what
            we desire already
    @param  en True to enable, False to disable
*/
/**************************************************************************/
/*
extern void TC0_Handler();
extern void TC1_Handler();
extern void TC2_Handler();
*/
extern void TC3_Handler();

void *timersThread(void *arg) {
	static int64_t startMicros, t, d, nbTicks=0;
	uint8_t timerNum=*(uint8_t *)arg;
	
	startMicros=getTime();
	
	//fprintf(stderr, "timersThread : timerNum:%d period:%u\n", timerNum, tcArray[timerNum].period);
	
	while(1) {
		++nbTicks;
		t=getTime();
		
		switch(timerNum) {
			//case 0 : TC0_Handler(); break;
			//case 1 : TC1_Handler(); break;
			//case 2 : TC2_Handler(); break;
			case 3 : TC3_Handler(); break;
		}
		
		d=startMicros+(nbTicks)*tcArray[timerNum].period-t;
		if(d<10) usleep(10); else usleep(d);
	}
	
	return NULL;
	
}


void Adafruit_ZeroTimer::enable(boolean en) {
  // First, check if its enabled
  //fprintf(stderr, "Adafruit_ZeroTimer::enable en:%d\n", en);
  
  if(tcArray[_timernum].tid) pthread_cancel(tcArray[_timernum].tid);
  
	if(en) {
		if(pthread_create(&tcArray[_timernum].tid, NULL, timersThread, &_timernum) != 0) {
			//fprintf(stderr," pthread_create timersThead error !\n");
			exit(1);
		}

		pthread_detach(tcArray[_timernum].tid);
	}

}

extern "C" {

// TODO: this could probably be optimized to be faster
static inline void __tc_cb_handler(uint32_t mask, uint32_t offset) {
}

/**************************************************************************/
/*!
    @brief  The function we call from within the IRQ function defined in the
   sketch. We can't have all the IRQ functions in this library because then no
   other library can use them. So we have a handler that the usercode must call.
    @param timerNum The timer we just got an IRQ for, 3 for TC3, 4 for TC4, etc!
*/
/**************************************************************************/
void Adafruit_ZeroTimer::timerHandler(uint8_t timerNum) {
	if(tcArray[timerNum].callback != NULL) tcArray[timerNum].callback();
}

};
