/*
	Copyright (C) 2022 Guénal Davalan. All right reserved.
	
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>

#include "log.h"
#include "Arduino.h"

#include "x11.h"
#include "wio-terminal.h"
#include "interrupt.h"


void maximizeUnmaximizeWindow(Display *display, Window win);
extern int usbConnect;
int bounce=1;

int lastPressState=-1;

#define NB_BUTTONS 9

typedef struct Rect {
	int x, y, w, h;
	Rect(int x, int y, int w, int h) : x(x), y(y), w(w), h(h) {}
} Rect;

struct Button {
	uint8_t ident;
	char name[10];
	Rect rect;
};

enum { A, B, C, UP, LEFT, RIGHT, DOWN, PRESS, RESET };
 
struct Button buttons[NB_BUTTONS] = {
	{ WIO_KEY_A, "A", { 1, 1, 1, 1 } },
	{ WIO_KEY_B, "B", { 1, 1, 1, 1 } },
	{ WIO_KEY_C, "C", { 1, 1, 1, 1 } },	
	{ WIO_5S_UP, "UP", { 1, 1, 1, 1 } },
	{ WIO_5S_LEFT, "LEFT", { 1, 1, 1, 1 } },	
	{ WIO_5S_RIGHT, "RIGHT", { 1, 1, 1, 1 } },
	{ WIO_5S_DOWN, "DOWN", { 1, 1, 1, 1 } },
	{ WIO_5S_PRESS, "PRESS", { 1, 1, 1, 1 } },
	{ WIO_RESET, "RESET", { 1, 1, 1, 1 } }
};

#ifdef __cplusplus
extern "C" {
#endif

static int testPin(int16_t pin) {
	if(pin >= NB_IO) { 
		fprintf(stderr, "[ERROR] digitlRead : pin=%u (max=%u) !\n", pin, NB_IO-1); 
		return 0; 
	}

	return 1;
}	


void pinMode(pin_size_t pin, uint8_t mode) {
	if(!testPin(pin)) return;
	if((pin==WIO_KEY_A || pin==WIO_KEY_B || pin==WIO_KEY_C || 
		pin==WIO_5S_UP || pin==WIO_5S_DOWN || pin==WIO_5S_LEFT ||
		pin==WIO_5S_RIGHT || pin==WIO_5S_PRESS) && mode != INPUT_NONE)
		cpu.ios[pin].setMode(INPUT_PULLUP);
	else 
		cpu.ios[pin].setMode(mode);
}

void digitalWrite(pin_size_t pin, uint8_t value) {
	if(!testPin(pin)) return;
	//fprintf(stderr, "digitalWrite(%d) : %d\n", pin, value?255:0);
	cpu.ios[pin].setValue(value?255:0);
			
	uint8_t last=(cpu.ios[pin].lastValue!=0);
	//LOW CHANGE RISING FALLING HIGH
	if(cpu.itVector[pin]!=NULL) {
		if(cpu.itMode[pin]==LOW && value==LOW) sendInterrupt(pin);
		else if (cpu.itMode[pin]==HIGH && value==HIGH) sendInterrupt(pin);
		else if (cpu.itMode[pin]==CHANGE && value!=last) sendInterrupt(pin);
		else if (cpu.itMode[pin]==RISING && value==1 && last==0) sendInterrupt(pin);
		else if (cpu.itMode[pin]==FALLING && value==0 && last==1) sendInterrupt(pin);
	}	
}

uint8_t digitalRead(pin_size_t pin) {
	if(!testPin(pin)) return 0;
	return cpu.ios[pin].getValue()?1:0;
}

#ifdef __cplusplus
}
#endif

int buttonNameToNum(char *name) {
	int i=0;
	while(i<NB_BUTTONS) {
		if(!strcmp(name, buttons[i].name)) return buttons[i].ident;
		++i;
	}
	return -1;
}

int buttonPinToNum(uint8_t pin) {
	int i=0;
	while(i<NB_BUTTONS) {
		if(buttons[i].ident == pin) return i;
		++i;
	}
	return -1;
}


uint16_t buttonPin(int key) {
	switch(key) {
		case XK_a      : return WIO_KEY_A;
		case XK_b      : return WIO_KEY_B;
		case XK_c      : return WIO_KEY_C;
		case XK_Return : return WIO_5S_PRESS;
		case XK_Up     : return WIO_5S_UP;
		case XK_Down   : return WIO_5S_DOWN;
		case XK_Left   : return WIO_5S_LEFT;
		case XK_Right  : return WIO_5S_RIGHT;
		default        : return 255;
	}
}

void whenKeyPress(KeySym key) {
	switch(key) {
					
		case XK_h   : 	if(getenv("HELP_CMD")) {
							fprintf(stderr, "%s\n", getenv("HELP_CMD"));  system(getenv("HELP_CMD")); 
						}
						break;
		case XK_r   :
		case XK_F5  : button(WIO_RESET, 1); break;
		case XK_F9  : button(WIO_COMPILE, 1); break;
		case XK_F11 : maximizeUnmaximizeWindow(disp, xwin); break;
		
		case XK_a:
		case XK_b:
		case XK_Return: if(!ctrl) button(buttonPin(key), 1); break;
		 
		case XK_c     : if(ctrl) quit(); else button(buttonPin(key), 1); break;
		case XK_Up    : if(ctrl) board.setOrientation(3); else button(buttonPin(key), 1); break;
		case XK_Down  : if(ctrl) board.setOrientation(1); else button(buttonPin(key), 1); break;
		case XK_Right : if(ctrl) board.setOrientation(board.getOrientation()-1); else button(buttonPin(key), 1); break;
		case XK_Left  : if(ctrl) board.setOrientation(board.getOrientation()+1); else button(buttonPin(key), 1); break;
							
	}
}


void quit() {
	if(getenv("MASTER_PID")) 
		kill(atoi(getenv("MASTER_PID")), SIGINT);
}


int intersection(int x, int y) {
    
    for(int i=0; i<NB_BUTTONS; ++i) {
		Rect r=buttons[i].rect;
		if(x>=r.x && x<=r.x+r.w && y>=r.y && y<=r.y+r.h) return buttons[i].ident;
	}
	
	return -1;
}


void defButton(int name, struct Rect rect) {
	memcpy(&buttons[name].rect, &rect, sizeof(rect));
} 

int lastBeginDone; // last state of display

double c;
Board board;

double scaleEnv;
int buzzer;

//pthread_mutex_t _drawLock = PTHREAD_MUTEX_INITIALIZER;
//pthread_mutex_t _drawLock = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;

void c565toRGB(unsigned c, int *r, int *g, int *b);

TFT_eSPI *display=NULL;

void button(uint16_t pin, uint8_t state) {
	
	// reset
	
	if(pin==WIO_RESET) {
		fprintf(stderr, "wio pid : %d ppid:%d\n", getpid(), atoi(getenv("MASTER_PID")));
		if(getenv("MASTER_PID"))  kill(atoi(getenv("MASTER_PID")), SIGUSR1);
		return;
	}
	
	if(pin==WIO_COMPILE) {
		fprintf(stderr, "wio pid : %d ppid:%d\n", getpid(), atoi(getenv("MASTER_PID")));
		if(getenv("MASTER_PID"))  kill(atoi(getenv("MASTER_PID")), SIGUSR2);
		return;
	}
	
	int v=cpu.ios[pin].value?1:0;
	
	// bounce
	
	int b=1;
	switch(bounce) {
		case 0  : b=0; break;
		case 1  : b=random(10);
				  //fprintf(stderr, "bounce : b:%d\n", b);
				  if(b>=2) b=0;
				  break;
		case 2  : b=1;
	}
	
	if(b) { 
		if(v==0) { // rebond seulement à l'appui, pas au relachement
			//fprintf(stderr, "*** rebond1:%u\n", millis());
			usleep(2000/speedup);
			digitalWrite(pin, !v); 
			usleep(2000/speedup);
			//fprintf(stderr, "*** rebond2:%u\n", millis());
			digitalWrite(pin, v);
			usleep(2000/speedup); 
		}
	}
	
	//fprintf(stderr, "digitalWrite : %d %d\n", pin, cpu.ios[pin].mode==INPUT_PULLUP?!state:state);
	
	//int s=cpu.ios[pin].mode==INPUT_PULLUP?!state:state;			
	//if(v==0) fprintf(stderr, "*** press:%u\n", millis());
	digitalWrite(pin, !v);
	//fprintf(stderr, "s:%d\n", s);
	
	board.layer0->dlock();
	
	int lock=0;
	if(display) {
		display->dlock();
		lock=1;
	}
	
	board.layer0->changeActive=0;
	
	if(state) board.drawButton(buttonPinToNum(pin), TFT_GREEN);
	else board.drawButton(buttonPinToNum(pin), TFT_MAGENTA);
	
	
	board.layer0->changeActive=1;
	
	board.drawFastLayer0();
	
	if(lock) display->dunlock();
	board.layer0->dunlock();
	
	
	
}


void buttonPressRelease(int x, int y, char state) {

	//static int oriX=0, oriY=0;

	if(state==1 && lastPressState==0) {
		//printf("ORI x: %d y: %d\n", x, y);
		//oriX=x; oriY=y;
	}

	if(state!=lastPressState) {
		lastPressState=state;
	}


	if(board.layer0->orientation&1) {
		x-=board.layer0->offX;
		y-=board.layer0->offY;
	}
	else {
		x-=board.layer0->offX;
		y-=board.layer0->offY;
	}
		
	x=x/board.scale;
	y=y/board.scale;
		
	int x1=0, y1=0;
	int w=board.layer0->_init_width;
	int h=board.layer0->_init_height;
	
	switch(board.layer0->orientation) {
		case 0 : x1=x; y1=y; break;
		case 1 : x1=w-y-1; y1=x; break;		
		case 2 : x1=w-x-1; y1=h-y-1; break;		
		case 3 : x1=y; y1=h-x-1; break;
	}
	
	int r=intersection(x1, y1);
		
	if(r!=-1) button(r, state);
}



Board::Board() {
	
	winWidth=winHeight=-1;
	
	c=320/485.0; // realDisplayWidth;
	border=20;
	
	width=(570+border*2)*c;
	height=(725+border*2)*c;
	
	displayOffX=60*c;
	displayOffY=120*c;
	
	char *v;
	int orientation;
	
	//if((v=getenv("SCALE"))==NULL) { fprintf(stderr, "Need SCALE env variable !\n"); exit(1); }
	setenv("SCALE", "1", 0);	
	v=getenv("SCALE");
	scaleEnv=atof(v);
	if(scaleEnv<=0) { fprintf(stderr, "Bad scale value (%s) !\n", v); exit(1); }
	
	board.scale=scaleEnv;
		
	//if((v=getenv("ORIENTATION"))==NULL) { fprintf(stderr, "Need ORIENTATION env variable !\n"); exit(1); }
	setenv("ORIENTATION", "3", 0);
	v=getenv("ORIENTATION");	
	orientation=atoi(v);
	
	layer0 = new TFT_eSPI(width,height,0);
	
	layer0->setRotation(0);	
	layer0->setOffsets(0,0);
	layer0->debug=0;
	layer0->begin();
	setOrientation(orientation);
	
	startX11Thread();
}

void Board::computeDisplayPositionInWindow() {
	display->offX=board.layer0->offX;
	display->offY=board.layer0->offY;
		
	int dx=0, dy=0;

	switch(board.layer0->orientation) {
		case 0: dx=-29; dy=0; break;
		case 1: dx=-40; dy=70; break;
		case 2: dx=29; dy=0;  break;
		case 3: dx=-40; dy=10; break;
	}
	
	display->offX=(winWidth-((display->_init_width)*scale))/2+dx*scale;
	display->offY=(winHeight-((display->_init_height)*scale))/2+dy*scale;	
}

void Board::computeBoardPositionInWindow() {
	double scale, xScale, yScale;
		
	xScale=winWidth*1.0/(board.width);
	yScale=winHeight*1.0/(board.height);
	
	
	if(board.layer0->orientation&1) {
		xScale=winWidth*1.0/(board.height);
		yScale=winHeight*1.0/(board.width);
	}
	
	
	scale=xScale;
	if(yScale<xScale) scale=yScale;
	
	scale=(int)scale;
	
	board.scale=scale;
	
	if(scale<1) scale=1;
	
	board.layer0->offX=(winWidth-(board.width*scale))/2;
	board.layer0->offY=(winHeight-(board.height*scale))/2;
	
	if(board.layer0->orientation&1) {
		board.layer0->offY=(winHeight-(board.width*scale))/2;
		board.layer0->offX=(winWidth-(board.height*scale))/2;
	}
	
	board.layer0->reset();
	
	if( ! (display != NULL && display->beginDone)) return;
	
}


void Board::setOrientation(int o) {
	o=o%4;
	if(o<0) o=3;
	layer0->orientation=o;
	if(display) display->orientation=o;
	computeBoardPositionInWindow();
	if(display && display->beginDone) computeDisplayPositionInWindow();
	buildLayer0();
}

int Board::getOrientation() {
	return layer0->orientation;
}

int getPixelColor(TFT_eSPI *tft, int x, int y) {
	
	if(tft == NULL|| !tft->beginDone) return -1;
		
	int w=tft->_init_width, h=tft->_init_height;
							
	int j=x, i=y;
				
	switch(tft->orientation) {
		case 0: x=j; y=i; break;
		case 1: x=w-i-1; y=j; break;
		case 2: x=h-j-1;  y=w-i-1; break;
		case 3: x=i; y=h-j-1; break;
	}
	
	int c=tft->frontBuffer[y*w+x];
	
	//tft->frontBuffer[y*w+x]=0xFFE0; tft->backBuffer[y*w+x]=1; tft->modified=1;

	/*	
	if(glogFlag && glogIndent==0) alog(std::string("getPixel(x:")+std::to_string(x)+", y:"+
		//std::to_string(y)+") "+int_to_hex((uint16_t)c));
	*/	
	return c;					
}

int getBlockColor(TFT_eSPI *tft, int x, int y, int w, int h) {
	
	if(glogFlag) glogInc();
	
	int c=getPixelColor(tft, x, y);
	
	for(int j=y; j<y+h; ++j) {
		for (int i=x; i<x+w; ++i) {		
			if(getPixelColor(tft, i, j) != c) {
				c=-1;
				break;
			}
		}
	}
	if(glogFlag) glogDec();
	
	/*
	if(glogFlag && glogIndent==0) alog(std::string("getPixel(x:")+std::to_string(x)+", y:"+
		std::to_string(y)+", w:"+
		std::to_string(w)+", h:"+
		//std::to_string(h)+") "+int_to_hex((uint16_t)c));
	*/
	return c;
}

int16_t getBlockCrc(TFT_eSPI *tft, int x, int y, int w, int h) {
	
	uint16_t crc=0;
	if(glogFlag) glogInc();
		
	for(int j=y; j<y+h; ++j) {
		for (int i=x; i<x+w; ++i) {		
			crc+=getPixelColor(tft, i, j);
			
		}
	}
	if(glogFlag) glogDec();
	
	/*
	if(glogFlag && glogIndent==0) alog(std::string("getPixel(x:")+std::to_string(x)+", y:"+
		std::to_string(y)+", w:"+
		std::to_string(w)+", h:"+
		//std::to_string(h)+") "+int_to_hex((uint16_t)c));
	*/
	return crc;
}

void drawScaled(TFT_eSPI *tft) {
	
	//fprintf(stderr, "drawScaled : %p\n", tft);
	
	if(tft == NULL|| !tft->beginDone) return;
		
	int w=tft->_init_width, h=tft->_init_height;
			
	int n=0;
	uint32_t cpt=0;
	
	// work around : force redraw all pixel
	//memset(tft->backBuffer, 1, w*h*sizeof(uint16_t));
	
	for(int j=0; j < h; ++j) {
		for(int i=0; i < w; ++i) {
				
			if(tft->backBuffer[n]) {
				cpt++;
				tft->backBuffer[n]=0;
				
				int x=j, y=i;
				
				switch(tft->orientation) {
					case 0: x=j; y=i; break;
					case 1: x=w-i-1; y=j; break;
					case 2: x=h-j-1;  y=w-i-1; break;
					case 3: x=i; y=h-j-1; break;
				}
				
				int c=tft->frontBuffer[n];				
				drawScalledPixel(tft, x, y, c);
				
			}
			
			n=n+1;
		}
	}


	
	//fprintf(stderr, "cpt:%d\n", cpt);
	
	tft->modified=0;
}

void Board::drawButton(int i, uint16_t color) {
	layer0->drawRect(buttons[i].rect.x+2, buttons[i].rect.y+2, buttons[i].rect.w-4, buttons[i].rect.h-4, color);
}

void Board::buildLayer0() {
	 
    double b=8.0;
    double r=40;
      
    int w=board.layer0->_width, h=board.layer0->_height;
    int ox=border*c, oy=border*c;
    
    // buttons A B C
    
    for(int i=0; i<3; ++i) {
		layer0->fillRoundRect(ox-3, oy+320*c+i*125*c, 5, 80*c, 10, TFT_BLUE);		
		defButton(i, Rect(ox-20*c , oy+320*c+i*125*c -3*c,  5 +18*c, 80*c + 6*c));
	}
	
	// off/on/reset
	layer0->fillRoundRect(ox+80, oy+h-36, 40*c, 20*c, 5, RGB(250, 250, 240));
	layer0->drawRoundRect(ox+80, oy+h-36, 40*c, 20*c, 5, TFT_LIGHTGREY);
	defButton(RESET, Rect(ox+80-5, oy+h-36+2, 40*c+10, 20*c+5));
    
    
    layer0->fillRoundRect(ox, oy, w-border*2*c, h-border*2*c, 18, RGB(250, 250, 240));    
    layer0->drawRoundRect(ox, oy, w-border*2*c, h-border*2*c, 18, TFT_LIGHTGREY);   
    layer0->fillRoundRect(ox+b*c, oy+b*c, 480*c, 720*c-2*b*c, 16, RGB(0,0,0));
    
    layer0->fillCircle(ox+490*c, oy+90*c, r*c, TFT_BLUE);
    defButton(PRESS, Rect(ox+490*c-7, oy+90*c-7, 14, 14));
        
    int t[4][8]= { 
		{ -1,  0,  0,  1, -1,  0,  0, -1 },
		{  0,  1, -1,  0,  0, 1,  1,  0 }, 
		{  0, -1, -1,  0,  0, -1,  1,  0 },
		{ 1,  0,  0,  1, 1,  0,  0, -1 },
		}; 
	
	for(int i=0; i<4; ++i) {
		int cx=ox+490*c + t[i][0]*16;
		int cy=oy+90*c  + t[i][1]*16;
		int l=3;

		layer0->fillTriangle(t[i][2]*l+cx, t[i][3]*l+cy, t[i][4]*l+cx, t[i][5]*l+cy, t[i][6]*l+cx, t[i][7]*l+cy, TFT_NAVY);
		defButton(UP+i, Rect(cx-7, cy-7, 14, 17));
	}
	
	for(int i=0; i<4; ++i) {
		layer0->fillRoundRect(ox+515*c, oy+570*c+i*20*c, 30*c, 10*c, 2, TFT_BLACK);
	}
	
    for(int i=0; i<NB_BUTTONS; ++i) {
		drawButton(i,TFT_MAGENTA); 
	}
}

void Board::drawLayer0() {
	layer0->dlock();
	layer0->modified=0;
	clearTftBackBuffer();
	layer0->reset();
	if(display!=NULL && display->beginDone) display->reset();
	drawScaled(layer0);
	display->modified=1;
	layer0->dunlock();
}

void Board::drawFastLayer0() {
	drawScaled(layer0);
}

void Board::drawDisplay() {
	
	if(lastBeginDone != display->beginDone) {
		computeDisplayPositionInWindow();
		lastBeginDone=display->beginDone;
	}
	
	drawScaled(display);
	display->modified=0;
}

void Board::draw(int windowSizeChanged, int forceRedraw) {
	
	if(windowSizeChanged||forceRedraw) {
		computeBoardPositionInWindow();
		if(display && display->beginDone) computeDisplayPositionInWindow();
	}
	
	
	if(forceRedraw || board.layer0->modified) {
		drawLayer0();		
	}
	
	if(display) display->orientation=layer0->orientation;
	
	if(display!=NULL  && display->beginDone && display->modified) {
		display->dlock();
		drawDisplay();	
		display->dunlock();	
	}

	copyBuffXwin();	

	/*
	if(doubleBuffered) {
		copyBuff();
		doubleBuffered=0;
		//forceRedraw=1;
	}
	*/

}

void *cmdThread(void *arg) {
	FILE *fifo;
	char line[200];
	char *fifoName=(char *)arg;
	
	fprintf(stderr, "Start cmdThread...\n");
	
	if((fifo=fopen(fifoName, "r"))==NULL) {
		perror("fifo fopen");
		exit(1);
	}
	
	while(fgets(line, 200, fifo) != NULL) {
		line[strlen(line)-1]=0;
		fprintf(stderr, "FIFO : %s\n", line);
		if(!strcmp(line, "reset")) {
			fprintf(stderr, "reset...\n");
			button(WIO_RESET, 1);
		}
		else if(!strncmp(line, "bounce ", 7)) {
			fprintf(stderr, "bounce : %s ...\n", line);
			bounce=atoi(line+7);
			fprintf(stderr, "bounce : %d\n", bounce);
		}
		else if(!strncmp(line, "usb ", 4)) {
			fprintf(stderr, "usb : %s ...\n", line);
			if(!strcmp(line+4,"connect")) usbConnect=1;
				else usbConnect=0;
			fprintf(stderr, "usb : %d\n", usbConnect);
		}
		else if(!strncmp(line, "click", 4)) {
			fprintf(stderr, "click : %s ...\n", line);
			char buttonName[20]; int delay1=200, delay2=100;
			sscanf(line, "click %s %d %d", buttonName, &delay1, &delay2);
			fprintf(stderr, "click : %s %d %d...\n", buttonName, delay1, delay2);
			int n=buttonNameToNum(buttonName);
			fprintf(stderr, "button num : %d\n", n);
			if(n!=-1) {
				button(n, 1);
				usleep(delay1*1000/speedup);
				button(n, 0);
				usleep(delay2*1000/speedup);
			}
		}
		else if(!strncmp(line, "getBlockColor", 13)) {
			int x, y, w, h;
			sscanf(line, "getBlockColor %d %d %d %d", &x, &y, &w, &h);
			int c=getBlockColor(display, x, y, w, h);
			if(glogFlag) {
				if (c==-1) fprintf(stderr, "getBlockColor %d %d %d %d %d\n", x, y, w, h, c);
				else fprintf(stderr, "getBlockColor %d %d %d %d 0x%04X\n", x, y, w, h, (uint16_t)c);
			}
		}
		else if(!strncmp(line, "getBlockCrc", 11)) {
			int x, y, w, h;
			sscanf(line, "getBlockCrc %d %d %d %d", &x, &y, &w, &h);
			uint16_t c=getBlockCrc(display, x, y, w, h);
			if(glogFlag) fprintf(stderr, "getBlockCrc %d %d %d %d 0x%X\n", x, y, w, h, c);
			
		}
		else {
			fprintf(stderr, "Invalid command from fifo : %s\n", line);
		}
	}
	
	return NULL;
}

void startCmdThread(char *fifoName) {
	pthread_t tid;
	if(pthread_create(&tid, NULL, cmdThread, (void *)fifoName) != 0) {
		fprintf(stderr," pthread_create cmdThead error !\n");
		exit(1);
	}

	pthread_detach(tid);

}

void initBoard() {
	char *fifoName;
	
	fprintf(stderr, "initBoard...\n");
	
	pinMode(WIO_KEY_A, INPUT_NONE);
	pinMode(WIO_KEY_B, INPUT_NONE);
	pinMode(WIO_KEY_C, INPUT_NONE);
	pinMode(WIO_5S_UP, INPUT_NONE);
	pinMode(WIO_5S_DOWN, INPUT_NONE);
	pinMode(WIO_5S_LEFT, INPUT_NONE);
	pinMode(WIO_5S_RIGHT, INPUT_NONE);
	pinMode(WIO_5S_PRESS, INPUT_NONE);
	
	
	if((fifoName=getenv("FIFO_CMD")) != NULL) {
		startCmdThread(fifoName);
	}
	fprintf(stderr, "initBoard done\n");
}
