/*	
	Copyright (C) 2022 Guénal Davalan. All right reserved.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>
#include <assert.h>

#include "x11.h"

extern void quit();

extern double scaleEnv;
int windowSizeChanged=1;
int forceRedraw;
int windowPresent;
int firstDraw;

//float accel_X=0.04, accel_Y=-0.01;

float accel_X_ori=0.04, accel_Y_ori=-0.01;

float accel_X=accel_X_ori, accel_Y=accel_Y_ori;

Window xwin;
Display *disp;
GC gc;
int ctrl=0;

int doubleBuffered=0;
XdbeBackBuffer xwinBuff;
XdbeBackBuffer doubleBuff;

int dragOriX=0;
int dragOriY=0;
int dragState=0;

struct Rect {
	int x, y;
};

static Rect rects[]= {
	{ 75, 0 },
	{ 159, 0 },
	{ 242, 0 },
	{ 412, 330 },
	{ 412, 305 },
	{ 412, 355 },
	{ 387, 330 },
	{ 437, 330 },
	{ -1, 0 }
};


GC create_gc(Display* display, Window win, int reverse_video)
{
	GC gc;				/* handle of newly created GC.  */
	unsigned long valuemask = 0;		/* which values in 'values' to  */
	/* check when creating the GC.  */
	XGCValues values;			/* initial values for the GC.   */


	unsigned int line_width = 2;		/* line width for the GC.       */
	int line_style = LineSolid;		/* style for lines drawing and  */
	int cap_style = CapButt;		/* style of the line's edje and */
	int join_style = JoinBevel;		/*  joined lines.		*/
	int screen_num = DefaultScreen(display);

	line_width=1;
	cap_style=CapNotLast;

	gc = XCreateGC(disp, xwin, valuemask, &values);
	
	/* allocate foreground and background colors for this GC. */
	if (reverse_video) {
		XSetForeground(display, gc, WhitePixel(display, screen_num));
		XSetBackground(display, gc, BlackPixel(display, screen_num));
	}
	else {
		XSetForeground(display, gc, BlackPixel(display, screen_num));
		XSetBackground(display, gc, WhitePixel(display, screen_num));
	}

	/* define the style of lines that will be drawn using this GC. */
	XSetLineAttributes(display, gc,
			line_width, line_style, cap_style, join_style);

	/* define the fill style for the GC. to be 'solid filling'. */
	XSetFillStyle(display, gc, FillSolid);

	return gc;
}

unsigned long c565toLong(unsigned c) {
    unsigned r,g,b;
    r=c >> 8; g=(c & 0x7E0) >> 3; b=(c & 0x1F) * 8;
    return b + (g<<8) + (r<<16);
}


int intersection(Rect *r, int x, int y) {
	int i=0;

	while(rects[i].x != -1) {	
		if(x>=rects[i].x && x<=rects[i].x+20 && y>=rects[i].y && y<=rects[i].y+20) break;
		++i;
	}

	if(rects[i].x != -1) return i;
	return -1;
}

int keycodeToButton(int keycode, int press) {
	int b=-1;

	switch(keycode) {
		// F1 : A
		case 0x43 : b=0; break;
					// F2 : B
		case 0x44 : b=1; break;
					// F3 : B
		case 0x45 : b=2; break;
					// PRESS 
		case 0x24 :
		case 0x68 : b=3; break;
					// UP
		case 0x6f : b=4; break;
					// DOWN
		case 0x74 : b=5; break;
					// LEFT
		case 0x71 : b=6; break;
					// RIGHT
		case 0x72 : b=7; break;
	}

	return b;
}


void updateWhenX11SizeChanged() {
	static int lastWinWidth=-1, lastWinHeight=-1;
		
	if(display!=NULL && display->beginDone) display->reset();
	
	XWindowAttributes winAttr;
	XGetWindowAttributes(disp,xwin,&winAttr);	
	
	windowSizeChanged=0;
	
	if(winAttr.width!=lastWinWidth || winAttr.height!=lastWinHeight) {
		
		board.winWidth=winAttr.width;
		board.winHeight=winAttr.height;
		
		windowSizeChanged=1;
		
		lastWinWidth=winAttr.width;
		lastWinHeight=winAttr.height;
	}	
}


void setX11ForegroundColor(uint32_t color) {
    static unsigned long last_foreground=0;
    unsigned long x11Color=0;

    if(color != last_foreground) {
        last_foreground=color;
        x11Color=c565toLong(color);
        
        XSetForeground(disp, gc, x11Color);
    }
}


void clearTftBackBuffer() {
	setX11ForegroundColor(0xffffff);
	XFillRectangle(disp, xwinBuff, gc, 0, 0, board.winWidth, board.winHeight);
}

    
void maximizeUnmaximizeWindow(Display *display, Window win) {
	static int state=0;
	
	XEvent ev;
	ev.xclient.window = win;
	ev.xclient.type = ClientMessage;
	ev.xclient.format = 32;
	ev.xclient.message_type = XInternAtom(display, "_NET_WM_STATE", False);
	ev.xclient.data.l[0] = 1;
	ev.xclient.data.l[1] = XInternAtom(display, "_NET_WM_STATE_MAXIMIZED_HORZ", False);
	ev.xclient.data.l[2] = XInternAtom(display, "_NET_WM_STATE_MAXIMIZED_VERT", False);
	ev.xclient.data.l[3] = 1;
		  
	if(state==0) 
		ev.xclient.data.l[0] = 1;
	else 
		ev.xclient.data.l[0] = 0;

	XSendEvent(display, DefaultRootWindow(display), False, SubstructureRedirectMask | SubstructureNotifyMask, &ev);
	state=!state;
	
	updateWhenX11SizeChanged();
}


void drawScalledPixel(TFT_eSPI *tft, int x, int y, uint16_t color) {
	
	long double scale=board.scale;

	float cx=1;
	float cy=1;

	//float refX, refY;

	//float ax=accel_X;
	//float ay=accel_Y;

	/*
	cx=ax/2+1;
	cy=ay/2+1;

	if(cx>1) cx=2-cx;
	if(cx<-1) cx=-2-cx;

	if(cy>1) cy=2-cy;
    if(cy<-1) cy=-2-cy;
	*/

	int sx=x*scale;
    int sy=y*scale;
	
	setX11ForegroundColor(color);

	if(doubleBuffered) {
		XFillRectangle(disp, doubleBuff, gc, tft->offX+sy, tft->offY+sx, scale, scale);
	}
	else {
		XFillRectangle(disp, xwinBuff, gc, tft->offX*cy, tft->offY*cx, scale, scale);
	}
}

void copyBuff() {
	//XCopyArea(disp, doubleBuff, xwin, gc, 0, 0, board.winWidth, board.winHeight,  0, 0);
}

void copyBuffXwin() {
    XCopyArea(disp, xwinBuff, xwin, gc, 0, 0, board.winWidth, board.winHeight,  0, 0);
}


void accelX(float v) {
	//printf("v: %f   accel_X : %f\n", v, accel_X);
	accel_X+=v;

}

void accelY(float v) {
	//printf("v: %f   accel_Y : %f\n", v, accel_Y);
	accel_Y+=v;
}


void getMousePosition(int *x, int *y) {

	static int lastX=-1, lastY=-1;

	int number_of_screens;
    //int i;
    Bool result;
    //Window *root_windows;
    Window window_returned;
    int root_x, root_y;
    int win_x, win_y;
    unsigned int mask_return;

    Display *display = disp; //XOpenDisplay(NULL);
    //assert(display);
    //XSetErrorHandler(_XlibErrorHandler);
    number_of_screens = XScreenCount(display);
    fprintf(stderr, "There are %d screens available in this X session\n", number_of_screens);

	result = XQueryPointer(display, xwin, &window_returned,
                &window_returned, &root_x, &root_y, &win_x, &win_y,
                &mask_return);

	if (result != True) {
        fprintf(stderr, "No mouse found.\n");
        //return -1;
    }
    else {
        *x=root_x; *y=root_y;
		if(lastX!=root_x || lastY != root_y) {
			if(dragState) {
				//printf("Mouse is at (%d,%d) rx:%d ry:%d  %d %d\n", root_x, root_y, dragOriX, dragOriY, root_x-dragOriX, root_y-dragOriY);
				float c=0.001;
				accel_Y=-(root_x-dragOriX)*c;
				accel_X=(root_y-dragOriY)*c;
			}
			lastX=root_x; lastY=root_y;
		}
	}

	
	
	/*
    root_windows = (Window *)malloc(sizeof(Window) * number_of_screens);
    for (i = 0; i < number_of_screens; i++) {
        root_windows[i] = XRootWindow(display, i);
    }
    for (i = 0; i < number_of_screens; i++) {
        result = XQueryPointer(display, root_windows[i], &window_returned,
                &window_returned, &root_x, &root_y, &win_x, &win_y,
                &mask_return);
        if (result == True) {
            break;
        }
    }
    if (result != True) {
        fprintf(stderr, "No mouse found.\n");
        //return -1;
    }
    else {
		*x=root_x; *y=root_y;
		printf("Mouse is at (%d,%d)\n", root_x, root_y);
	}

    free(root_windows);
	*/


}

void updateDrag() {

	extern int lastPressState;
	if(lastPressState==1) {
		int x,y;
		getMousePosition(&x, &y);
	}

}


void *x11Thread(void *arg) {
	unsigned long cpt=0;
	
	arg=(int *) arg; // do nothing, to avoid warning
	
	int stop=0;
	
	XInitThreads();
	
	/* open the display (connect to the X server) */

	disp = XOpenDisplay(NULL);
	if (disp == NULL)
	{
		fprintf(stderr, "Cannot open display\n");
		exit(1);
	}
    
    int s = DefaultScreen(disp);
	int max=board.width;
	if(board.height>max) max=board.height;
	
	xwin = XCreateSimpleWindow(disp, RootWindow(disp, s), 0, 0, 
	max*scaleEnv, max*scaleEnv, 1, BlackPixel(disp, s), WhitePixel(disp, s));
	
	int major_version_return, minor_version_return;
    if(!XdbeQueryExtension(disp, &major_version_return, &minor_version_return)) {
        fprintf(stderr, "XDBE is not supported!!!1\n");
        exit(1);
    }
	
	xwinBuff=XdbeAllocateBackBufferName(disp, xwin, 0);
	doubleBuff=XdbeAllocateBackBufferName(disp, xwin, 0);
	
	XMapWindow(disp, xwin);
	
	if(getenv("GEOMETRY") != NULL) {
		int x,y;
		unsigned w,h;
		XParseGeometry(getenv("GEOMETRY"), &x, &y, &w, &h);
		XMoveResizeWindow(disp, xwin, x, y, w, h);
	}
	
	if(getenv("NAME")) 
		XStoreName(disp, xwin, getenv("NAME")); 
	
    XSizeHints *wind_size_hints;
    
    if (!(wind_size_hints = XAllocSizeHints()))
	{
		fprintf(stderr, "Could not allocate memory for size hints.\n");
		exit(1);
	}
	
	wind_size_hints->flags = PMinSize;	

	wind_size_hints->min_width = max;
	wind_size_hints->min_height = max;

	XSetWMNormalHints(disp, xwin, wind_size_hints);

	XFree(wind_size_hints);
	    
    gc = create_gc(disp, xwin, 0);
    	
	/* select kind of events we are interested in */
	XSelectInput(disp, xwin, KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | ExposureMask);
	
	// register interest in the delete window message
	Atom wmDeleteMessage = XInternAtom(disp, "WM_DELETE_WINDOW", False);
	XSetWMProtocols(disp, xwin, &wmDeleteMessage, 1);
	
	char buffer[20];
	int bufsize = 20;
	KeySym key;
	XComposeStatus compose;
	
	int fps=getEnvInt((char *)"FPS", 25);
	uint32_t period=1000000L/fps;
	
	fprintf(stderr, "fps:%d, period:%u\n", fps, period);
	static uint64_t startMicros, t, nbTicks=0;	
	int64_t d;
	startMicros=getTime();

	float lx=0, ly=0;
	
	while (!stop)
	{
		
		XEvent event;

		while (XPending(disp)) {
			XNextEvent(disp, &event);
			
			if (event.type == Expose) {
					
				updateWhenX11SizeChanged();
				windowPresent=1;
				doubleBuffered=1;
				forceRedraw=1;

			}

			if(lx!=accel_X || ly!=accel_Y) {
				//printf("accel x: %f y: %f\n", accel_X, accel_Y);
				lx=accel_X; ly=accel_Y;
				forceRedraw=1;
			}   


			/* keyboard events */
			else if (event.type == KeyPress)
			{
				/* exit on ESC key press */
				if ( event.xkey.keycode == 0x09 ) 
					stop=1;
					
				ctrl=(event.xkey.state & ControlMask);
					
				XLookupString((XKeyEvent*)&event, buffer, bufsize, &key, &compose);				
				
				whenKeyPress(key);
			}
			else if (event.type == KeyRelease)
			{
				unsigned short is_retriggered = 0;

				if (XEventsQueued(disp, QueuedAfterReading))
				{
					XEvent nev;
					XPeekEvent(disp, &nev);

					if (nev.type == KeyPress && nev.xkey.time == event.xkey.time &&
							nev.xkey.keycode == event.xkey.keycode)
					{
						// delete retriggered KeyPress event
						XNextEvent (disp, &event);
						is_retriggered = 1;
					}
				}

				if (!is_retriggered) {

					//float c=0.01;
					
					XLookupString((XKeyEvent*)&event, buffer, bufsize, &key, &compose);
					
					switch(key) {
						case XK_a      : if(!ctrl) button(buttonPin(key), 0); break;
						case XK_b      : if(!ctrl) button(buttonPin(key), 0); break;
						case XK_c      : if(!ctrl) button(buttonPin(key), 0); break;
						case XK_Return : if(!ctrl) button(buttonPin(key), 0); break;
						//case XK_Up     : accelX(-c); if(!ctrl) button(buttonPin(key), 0); break;
						case XK_Up     : if(!ctrl) button(buttonPin(key), 0); break;
						//case XK_Down   : accelX(c); if(!ctrl) button(buttonPin(key), 0); break;
						case XK_Down   : if(!ctrl) button(buttonPin(key), 0); break;
						//case XK_Left   : accelY(c); if(!ctrl) button(buttonPin(key), 0); break;
						case XK_Left   : if(!ctrl) button(buttonPin(key), 0); break;
						//case XK_Right  : accelY(-c); if(!ctrl) button(buttonPin(key), 0); break;
						case XK_Right  : if(!ctrl) button(buttonPin(key), 0); break;
					}
				}
			}

			// buttons

			else if (event.type==ButtonPress) {
				int x, y;
				getMousePosition(&x, &y);
				//printf("Set : %d %d\n", x, y);
				dragOriX=x; dragOriY=y;
				dragState=1;
				buttonPressRelease(event.xbutton.x, event.xbutton.y, 1);
			}

			else if (event.type==ButtonRelease) {
				dragState=0;
				buttonPressRelease(event.xbutton.x, event.xbutton.y, 0);
				// reset accel
				accel_X=accel_X_ori; accel_Y=accel_Y_ori;
				forceRedraw=1;
				
			}

			// other

			else if (event.type == ClientMessage &&
					(unsigned long) event.xclient.data.l[0] == wmDeleteMessage) {
					stop=1;
				break;
			}	


			// Mouse (drag)
			if(dragState) {
				int x, y;
				getMousePosition(&x, &y);
			}
		}

		
		if(!XPending(disp)) {		
			board.draw(windowSizeChanged, forceRedraw);
			
			XFlush(disp);
			XSync(disp, False);
			
			if(windowSizeChanged) usleep(10000); // work around : bold
			
			windowSizeChanged=forceRedraw=0;
			
		}

		//updateDrag();
		
		/*work around : bold
		if(!firstDraw) {
			usleep(10000);
			updateWhenX11SizeChanged(); forceRedraw=1;
			firstDraw=1;
		}
		*/
		
		XFlush(disp);
		XSync(disp, False);
		
		
		++nbTicks;
		t=getTime();
		//fprintf(stderr, "s:%lu delta=%lu t:%lu\n", startMicros, nbTicks*period, t);
		d=(startMicros+nbTicks*period)-t;
		///speedup;
		//fprintf(stderr, "d:%ld\n", d);
		
		//usleep(period);
		if(d>0) usleep(d);
				
		
		//fprintf(stderr, "frame...\n");
				
		cpt++;		
	}

	XCloseDisplay (disp);
	quit();
	return 0;	
}

void startX11Thread() {
	pthread_t tid;
	if(pthread_create(&tid, NULL, x11Thread, NULL) != 0) {
		fprintf(stderr," pthread_create x11Thead error !\n");
		exit(1);
	}

	pthread_detach(tid);

}













